/*
 * YM player for the 32X
 * Mic, 2013
 */

#include <sys/types.h>
#include <string.h>
#include <stdbool.h>
#include <malloc.h>
#include "32x.h"
#include "hw_32x.h"
#include "ym_emu.h"
#include "sound.h"

#define VISUALIZER_CENTER_Y 42
#define VISUALIZER_X 204
#define VISUALIZER_STEPS 35
#define VISUALIZER_OFFS_INC ((SAMPLE_RATE/50)/VISUALIZER_STEPS)

#define CHN_INDICATOR_X 268
#define CHN_INDICATOR_Y 184
#define CHN_INDICATOR_X_STRIDE 6
#define CHN_INDICATOR_Y_STRIDE 3

#define SONG_INFO_X 59
#define SONG_INFO_Y 176

#define SONG_LIST_X 68
#define SONG_LIST_Y 75

#define PLAYTIME_X 262
#define PLAYTIME_Y 19

#define WHITE_LINE_X 43
#define WHITE_LINE_Y 74
#define BLACK_LINE_X 43
#define BLACK_LINE_Y 72

#define TOTAL_SONGS 30
#define SONGS_TO_SHOW 10
#define SCROLLBAR_Y 83
#define SCROLLBAR_X 261
#define SCROLLBAR_HEIGHT (64*SONGS_TO_SHOW/TOTAL_SONGS)


// From songs/songs.s
extern const uint8_t *SONG_POINTERS[];
extern const size_t SONG_LENGTHS[];
extern const uint8_t *SONG_FILENAMES[];
extern const size_t NUM_SONGS;
extern const size_t TOTAL_SONG_SIZE;

// From gfx/gfx.s
extern const uint16_t PALETTE_DATA[];
extern const uint8_t BACKGROUND_GFX[];
extern const uint8_t SMALL_FONT[];

// Holds Y-coordinates for the scrollbar
uint16_t scrollBarTable[TOTAL_SONGS+1-SONGS_TO_SHOW];

size_t currSong = 14, highlightedSong = 14, firstShownSong = 12;
uint32_t playCmd = CMD_IDLE;
volatile uint32_t *playCmdCacheThrough = &playCmd;

char listHeaderString[] = "1000000 bytes in 00 items.";
char songLengths[TOTAL_SONGS][6];

extern uint32_t soundTimeMs;
volatile uint32_t *soundTimeMsCacheThrough = &soundTimeMs;
char playTimeText[] = "00:00";

int16_t *sampleBuffer = NULL;
volatile int16_t **sampleBufferCacheThrough;

uint32_t filterFactor = 216, filterStep = 0, filterInc;

int16_t __attribute__((aligned(16))) visualizerData[(SAMPLE_RATE/50)];
uint8_t visualizerBitmap[70*20];

volatile uint16_t *frameBuffer16;
volatile uint16_t *cram16;
volatile uint16_t *overwriteImg16;

volatile bool showNfo = false;

// From ym_emu.c
extern int8_t	modeA, modeB, modeC;
extern int8_t	toneA, toneB, toneC;
extern int8_t   noiseA, noiseB, noiseC;
extern ymEffect	effects[2];


/*
 * Put a string on the overwrite image using the small font
 */
static void small_font_puts(const char *str, int x, int y, uint32_t baseCol, int delim)
{
	int c;
	uint8_t *fb8 = overwriteImg16 + 0x100;
	uint8_t pix;
	int screenOffs;
	int fontOffs;

	for (int i = 0; i < 23; i++) {
		c = str[i];
		if (c==delim || !c) break;

		screenOffs = y*320 + i * 8 + x;
		fontOffs = (c >> 4) << 10;
		fontOffs += ((c & 15) << 3);
		for (int t = 0; t < 8; t++) {
			for (int s = 0; s < 7; s++) {
				pix = SMALL_FONT[fontOffs + s] + baseCol;
				if (pix) pix += 128;
				if (pix)
					fb8[screenOffs + s] = pix;
			}
			screenOffs += 320;
			fontOffs += 128;
		}
	}
}


/*
 * Erase the previously drawn song info
 */
static void hide_song_info()
{
	int screenOffs = 0x100 + ((SONG_INFO_Y*320 + SONG_INFO_X) >> 1);
	uint16_t *p = &frameBuffer16[screenOffs];

	for (int i = 0; i < 24; i++) {
		memset(p, 0x80, 23*8);	// 0x80 is the background color for the song info box
		p += 160;
	}
}


/*
 * Refresh the song info "box" (title, author, copyright, sug-song)
 */
static void show_song_info()
{
	hide_song_info();

	small_font_puts(ymSongTitle,   SONG_INFO_X, SONG_INFO_Y, 0, '');
	small_font_puts(ymSongAuthor,  SONG_INFO_X, SONG_INFO_Y+8, 0,  '');
	small_font_puts(ymSongComment, SONG_INFO_X, SONG_INFO_Y+16, 0, '');
}


/*
 * Refresh the song list
 */
static void puts_active_list()
{
	int y = SONG_LIST_Y;

	for (int i = firstShownSong; i < firstShownSong+SONGS_TO_SHOW; i++) {
		char *fname = SONG_FILENAMES[i];

		if (i == highlightedSong) {
			for (int k = 0; k < 8; k++) {
				memset(frameBuffer16 + 0x100 + (y+k)*160 + (SONG_LIST_X/2),
					129,
					160);
			}
			small_font_puts(fname, SONG_LIST_X, y, 4, '.');
			small_font_puts("YM", SONG_LIST_X+72, y, 4, '');
			small_font_puts(songLengths[i], SONG_LIST_X+96, y, 4, '');
			small_font_puts((i < 20) ? "ST" : "ZX", SONG_LIST_X+144, y, 4, '');
		} else {
			for (int k = 0; k < 8; k++) {
				memset(frameBuffer16 + 0x100 + (y+k)*160 + (SONG_LIST_X/2),
					128,
					160);
			}
			small_font_puts(fname, SONG_LIST_X, y, 0, '.');
			small_font_puts("YM", SONG_LIST_X+72, y, 0, '');
			small_font_puts(songLengths[i], SONG_LIST_X+96, y, 0, '');
			small_font_puts((i < 20) ? "ST" : "ZX", SONG_LIST_X+144, y, 0, '');
		}
		y += 8;
	}
}


/*
 * Show the current play time in "MM:SS"
 */
static void show_play_time()
{
	for (int i = 0; i < 8; i++) {
		memset(frameBuffer16 + 0x100 + (PLAYTIME_Y+i)*160 + (PLAYTIME_X >> 1),
			0x80,
			5*8);
	}

	uint32_t timeMs = *soundTimeMsCacheThrough;
	uint32_t minutes = timeMs / (60000<<8);

	if (minutes >= 100) {
		minutes = 0;
	}
	uint32_t seconds = (timeMs - minutes * (60000<<8)) / (1000<<8);

	playTimeText[0] = (minutes / 10) + '0';
	playTimeText[1] = (minutes % 10) + '0';
	playTimeText[3] = (seconds / 10) + '0';
	playTimeText[4] = (seconds % 10) + '0';

	small_font_puts(playTimeText, PLAYTIME_X, PLAYTIME_Y, 0, '.');
}


/*
 * Draw the bar indicating where in the song list we're located
 */
static void update_scroll_bar()
{
	uint8_t *fb8 = frameBuffer16 + 0x100 + SCROLLBAR_Y*160;
	uint8_t *src;
	int barStart = scrollBarTable[firstShownSong];
	int barEnd = barStart+SCROLLBAR_HEIGHT-1;

	fb8 += SCROLLBAR_X;

	for (int y = 0; y < 64; y++) {
		if ((y+SCROLLBAR_Y) == barStart || (y+SCROLLBAR_Y) == barEnd) {
			src = BACKGROUND_GFX + BLACK_LINE_Y*320 + BLACK_LINE_Y;

		} else if ((y+SCROLLBAR_Y) >= barStart && (y+SCROLLBAR_Y) <= barEnd) {
			src = BACKGROUND_GFX + WHITE_LINE_Y*320 + WHITE_LINE_Y;

		} else {
			src = BACKGROUND_GFX + (y+SCROLLBAR_Y)*320 + SCROLLBAR_X;
		}
		memcpy(fb8, src, 10);
		fb8 += 320;
	}
}


/*
 * Draw the 3x5 grid of status indicators for each of the three YM/AY channels
 */
void draw_channel_inidcators(int x, int y, int8_t toneOff, int8_t noiseOff, int8_t usesEG, int8_t usesSFX1, int8_t usesSFX2)
{
	uint8_t *fb8 = frameBuffer16 + 0x100 + y*160;
	uint8_t c;

	fb8 += x;

	c = 128 + (toneOff^1);
	fb8[0] = c; fb8[1] = c; fb8[2] = c;
	fb8[320] = c; fb8[321] = c; fb8[322] = c;
	fb8 += CHN_INDICATOR_Y_STRIDE * 320;

	c = 128 + (noiseOff^1);
	fb8[0] = c; fb8[1] = c; fb8[2] = c;
	fb8[320] = c; fb8[321] = c; fb8[322] = c;
	fb8 += CHN_INDICATOR_Y_STRIDE * 320;

	c = (usesEG ? 129 : 128);
	fb8[0] = c; fb8[1] = c; fb8[2] = c;
	fb8[320] = c; fb8[321] = c; fb8[322] = c;
	fb8 += CHN_INDICATOR_Y_STRIDE * 320;

	c = (usesSFX1 ? 129 : 128);
	fb8[0] = c; fb8[1] = c; fb8[2] = c;
	fb8[320] = c; fb8[321] = c; fb8[322] = c;
	fb8 += CHN_INDICATOR_Y_STRIDE * 320;

	c = (usesSFX2 ? 129 : 128);
	fb8[0] = c; fb8[1] = c; fb8[2] = c;
	fb8[320] = c; fb8[321] = c; fb8[322] = c;
	fb8 += CHN_INDICATOR_Y_STRIDE * 320;
}


static void draw_all_channel_indicators()
{
	draw_channel_inidcators(CHN_INDICATOR_X, CHN_INDICATOR_Y, toneA, noiseA, modeA, (effects[0].channel==1), (effects[1].channel==1));
	draw_channel_inidcators(CHN_INDICATOR_X + CHN_INDICATOR_X_STRIDE, CHN_INDICATOR_Y, toneB, noiseB, modeB, (effects[0].channel==2), (effects[1].channel==2));
	draw_channel_inidcators(CHN_INDICATOR_X + CHN_INDICATOR_X_STRIDE*2, CHN_INDICATOR_Y, toneC, noiseC, modeC, (effects[0].channel==3), (effects[1].channel==3));
}


/*
 * Draw a line on the visualizer bitmap
 */
static void draw_line(int x0, int y0, int x1, int y1)
{
	y0 -= 27;
	y0 = (y0 < 0) ? 0 : y0;
	y1 -= 27;
	y1 = (y1 < 0) ? 0 : y1;

	x0 -= VISUALIZER_X;
	x1 -= VISUALIZER_X;

	int dx = (x1-x0), sx = x0<x1 ? 1 : -1;
	int dy = (y1-y0), sy = y0<y1 ? 1 : -1;

	if (dx < 0) dx = -dx;
	if (dy < 0) dy = -dy;

	int err = (dx>dy ? dx : -dy)>>1, e2;

	for(;;) {
		visualizerBitmap[y0*70+x0] = 129;
		if (x0==x1 && y0==y1) break;
		e2 = err;
		if (e2 >-dx) { err -= dy; x0 += sx; }
		if (e2 < dy) { err += dx; y0 += sy; }
	}
}


/*
 * Plot parts of the current sample buffer.
 * The waveform is drawn into an offscreen bitmap, which then
 * can be displayed as many times as needed.
 */
static void plot_waveform()
{
	static uint16_t lastSample = 6;
	static int oddFrame = 1;

	oddFrame ^= 1;
	if (oddFrame) return;

	memset(visualizerBitmap, 128, 70*20);

	int offs = 0;
	int x = VISUALIZER_X;
	int nextx = x+4;

	if (*sampleBufferCacheThrough) {
		for (int i = 0; i < (VISUALIZER_STEPS/2); i++) {
			uint16_t smp = visualizerData[offs] - 236;	// 236 = sample offset
			smp >>= 4;
			offs += VISUALIZER_OFFS_INC;
			draw_line(x, VISUALIZER_CENTER_Y-lastSample, nextx-1, VISUALIZER_CENTER_Y-smp);
			lastSample = smp;

			x = nextx;
			if (i < (VISUALIZER_STEPS/2)-2)
				nextx += 4;
			else
				nextx += 6;
		}
	}
}


/*
 * Copy the visualizer bitmap to the framebuffer
 */
static void show_waveform()
{
	for (int i = 0; i < 19; i++) {
		memcpy(frameBuffer16 + 0x100 + (VISUALIZER_CENTER_Y+i-9)*160 + (VISUALIZER_X/2),
			visualizerBitmap + i*70,
			VISUALIZER_STEPS*2);
	}
}


/*
 * Update both framebuffers
 */
static void update_screen()
{
	Hw32xScreenFlip(1);
	puts_active_list();
	show_song_info();
	update_scroll_bar();
	show_waveform();
	draw_all_channel_indicators();

	Hw32xScreenFlip(1);
	puts_active_list();
	show_song_info();
	update_scroll_bar();
	show_waveform();
	draw_all_channel_indicators();
}


/*
 * Start playback of a new song
 */
static void new_song(int index)
{
    // lock the mixer while killing the old music
    LockMixer(MIXER_LOCK_MSH2);

    UnlockMixer();

    Hw32xDelay(1); // allow the mixer to see the music is dead

    // lock the mixer while starting the new music
    LockMixer(MIXER_LOCK_MSH2);

	*playCmdCacheThrough = (CMD_SWITCH_SONG | index);
	*soundTimeMsCacheThrough = 0;

    UnlockMixer();

    update_screen();
}



/*
 * Move to the next entry in the song list
 */
static void move_to_next_list_item()
{
	if (highlightedSong < (NUM_SONGS - 1)) {
		highlightedSong++;
	}

	// Check if the song list needs to be scrolled
	if ( ((highlightedSong - firstShownSong) >= ((SONGS_TO_SHOW >> 1) + 1)) &&
		 (firstShownSong  < (NUM_SONGS - SONGS_TO_SHOW)) )
		firstShownSong++;

	update_screen();
}


/*
 * Move to the previous entry in the song list
 */
static void move_to_previous_list_item()
{
	if (highlightedSong > 0) {
		highlightedSong--;
	}

	// Check if the song list needs to be scrolled
	if ( (firstShownSong > highlightedSong) ||
		 ((firstShownSong) &&
		 ((highlightedSong - firstShownSong) < (SONGS_TO_SHOW >> 1))) )
		firstShownSong--;

	update_screen();
}



int main( void )
{
    unsigned short new_buttons, curr_buttons;
    unsigned short buttons = 0, paused = 0;
	showNfo = false;

	for (int i = 0; i < (SAMPLE_RATE/50); i++) {
		visualizerData[i] = 236;
	}

	// Use cache-through addresses for these pointers so that data can
	// be passed between the master and slave CPUs
	playCmdCacheThrough      = (volatile uint32_t*)((uint32_t)&playCmd      | 0x20000000);
	soundTimeMsCacheThrough  = (volatile uint32_t*)((uint32_t)&soundTimeMs  | 0x20000000);
	sampleBufferCacheThrough = (volatile int16_t**)((uint32_t)&sampleBuffer | 0x20000000);

    Hw32xInit(MARS_VDP_MODE_256);

	frameBuffer16 = &MARS_FRAMEBUFFER;
	overwriteImg16 = &MARS_OVERWRITE_IMG;
	cram16 = &MARS_CRAM;

	for (int i = 0; i < 200+2+9+9+8+2; i++)	{
		cram16[i] = PALETTE_DATA[i] & 0x7FFF;
	}
	cram16[128] = 0x7FFF;	// Set color 128 to white
	cram16[129] = 0;		// Set color 129 to black
	cram16[128+4] = 0;		// Set color 132 to black
	cram16[128+5] = 0x7FFF;	// Set color 133 to white


	Hw32xScreenFlip(1);
	memcpy(frameBuffer16 + 0x100, BACKGROUND_GFX, 320*224);
	Hw32xScreenFlip(1);
	memcpy(frameBuffer16 + 0x100, BACKGROUND_GFX, 320*224);
	small_font_puts("Initializing...", SONG_INFO_X, SONG_INFO_Y, 0, '');
	Hw32xScreenFlip(1);

	// Prepare the song length strings ("MM:SS")
	for (int i = 0; i < TOTAL_SONGS; i++) {
		uint8_t *p = SONG_POINTERS[i];
		uint32_t numFrames = (uint32_t)p[0x0E] << 8;
		numFrames += p[0x0F];
		numFrames /= 50;
		sprintf(songLengths[i], "%02u:%02u", (unsigned int)numFrames/60, (unsigned int)numFrames%60);
	}

	uint32_t scrollBarY = SCROLLBAR_Y << 8;
	for (int i = 0; i < TOTAL_SONGS+1-SONGS_TO_SHOW; i++) {
		scrollBarTable[i] = scrollBarY >> 8;
		scrollBarY += ((64 - SCROLLBAR_HEIGHT)*256) / (TOTAL_SONGS - SONGS_TO_SHOW);
	}
	scrollBarTable[TOTAL_SONGS - SONGS_TO_SHOW]++;

    while (MARS_SYS_COMM6 != MIXER_UNLOCKED) ; // wait for sound subsystem to init

	sprintf(listHeaderString, "%d bytes in %d ite", TOTAL_SONG_SIZE, NUM_SONGS);
	small_font_puts(listHeaderString, 44, 64, 0, '');
	small_font_puts("ms.", 44+23*8, 64, 0, '');
	Hw32xScreenFlip(1);
	small_font_puts(listHeaderString, 44, 64, 0, '');
	small_font_puts("ms.", 44+23*8, 64, 0, '');
	Hw32xScreenFlip(1);

	new_song(currSong);

    while (1)
    {
		memcpy(visualizerData, *sampleBufferCacheThrough, (SAMPLE_RATE/50)*2);
		plot_waveform();
		show_waveform();
		draw_all_channel_indicators();
		show_play_time();
		Hw32xScreenFlip(1);

        Hw32xDelay(1);

        // MARS_SYS_COMM10 holds the current button values: - - - - M X Y Z S A C B R L D U
        curr_buttons = MARS_SYS_COMM8;
        if ((curr_buttons & SEGA_CTRL_TYPE) == SEGA_CTRL_NONE)
            curr_buttons = MARS_SYS_COMM10; // if no pad 1, try using pad 2

        new_buttons = curr_buttons ^ buttons; // set if button changed
        buttons = curr_buttons;

        if (new_buttons & SEGA_CTRL_START) {
            if (!(buttons & SEGA_CTRL_START)) {
                // START just released
                showNfo = !showNfo;
                MARS_SYS_COMM0 = 0x4FF;		// Tell the 68k to show/hide the NFO text
            }
        }

        if (new_buttons & SEGA_CTRL_A) {
            if (!(buttons & SEGA_CTRL_A)) {

 			}
		}

        if (new_buttons & SEGA_CTRL_C) {
            if (!(buttons & SEGA_CTRL_C)) {

 			}
		}

        if (new_buttons & SEGA_CTRL_B) {
            if (!(buttons & SEGA_CTRL_B)) {
				if (!showNfo) {
					currSong = highlightedSong;
					new_song(currSong);
				}
 			}
		}

        if (new_buttons & SEGA_CTRL_DOWN) {
            if (!(buttons & SEGA_CTRL_DOWN)) {
                // DOWN just released
                if (!showNfo) move_to_next_list_item();
            }
        }

        if (new_buttons & SEGA_CTRL_UP) {
            if (!(buttons & SEGA_CTRL_UP)) {
                // UP just released
                if (!showNfo) move_to_previous_list_item();
			}
        }

        if (new_buttons & SEGA_CTRL_RIGHT) {
            if (!(buttons & SEGA_CTRL_RIGHT)) {
                // RIGHT just released
            }
        }

        if (new_buttons & SEGA_CTRL_LEFT) {
            if (!(buttons & SEGA_CTRL_LEFT)) {
                // LEFT just released
            }
        }

    }


    LockMixer(MIXER_LOCK_MSH2); // locked - stop playing

    return 0;
}

