// YM emulator for the 32X
// Mic, 2013

#include "32x.h"
#include "sound.h"
#include "ym_emu.h"

#define YM_MIXER		7
#define YM_LEVEL_A		8
#define YM_LEVEL_B		9
#define YM_LEVEL_C		10
#define YM_ENVE_FREQL	11
#define YM_ENVE_FREQH	12
#define YM_ENVE_SHAPE	13


static const uint16_t YM_VOL_TB[] = {
	// for (i : 0..15) YM_VOL_TB[i] = floor(power(10, (i-15)/6.67)*82)
	0x0000,0x0000,0x0000,0x0001,
	0x0001,0x0002,0x0003,0x0005,
	0x0007,0x000A,0x000E,0x0014,
	0x001D,0x0029,0x003A,0x0052
};

static const uint16_t YM2149_VOL_TB[] = {
	0x0000,0x0000,0x0000,0x0000,
	0x0000,0x0000,0x0000,0x0001,
	0x0001,0x0001,0x0001,0x0002,
	0x0002,0x0003,0x0003,0x0004,
	0x0005,0x0006,0x0007,0x0008,
	0x0009,0x000B,0x000C,0x000E,
	0x0011,0x0013,0x0015,0x0019,
	0x001D,0x0029,0x003A,0x0052
};

static const uint32_t TIMER_PRESCALER_TB[] = {
	0,4*512,10*512,16*512,
	50*512,64*512,100*512,200*512
};

static const uint32_t YM_DIGIDRUM_8TO4_TB[] = {
	0,0,1,1,3,4,4,5,5,5,6,6,6,6,7,7,
	7,7,7,7,8,8,8,8,8,8,8,8,9,9,9,9,
	9,9,9,9,9,9,9,10,10,10,10,10,10,10,10,10,
	10,10,10,10,10,10,10,11,11,11,11,11,11,11,11,11,
	11,11,11,11,11,11,11,11,11,11,11,11,11,11,12,12,
	12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,12,
	12,12,12,12,12,12,12,12,12,12,12,12,12,13,13,13,
	13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,
	13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,13,
	13,13,13,13,13,13,13,13,13,13,14,14,14,14,14,14,
	14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,
	14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,
	14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,14,
	14,14,14,14,14,14,14,14,14,14,15,15,15,15,15,15,
	15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,
	15,15,15,15,15,15,15,15,15,15,15,15,15,15,15,15
};

static uint8_t	 *ymData;
static uint8_t   *digiDrumPtr[32];
static uint32_t  digiDrumLen[16];
static int16_t	 digiDrumOverride[3];
static uint8_t   *ymRegStream;
static uint16_t  *ymEnvTable;
static uint8_t	 ymRegs[16];
static uint16_t  newRegValues;
static uint32_t  frame, numFrames, loopFrame, loopOffs, dataOffs;
static uint32_t  digiDrumFormat;
static uint16_t  numDigiDrums;
ymEffect  effects[2];

static uint32_t  posA, posB, posC;
static uint32_t  posE, posN;
static uint32_t  speedTN, speedE, speedT1, speedT2;
static uint32_t  periodA, periodB, periodC, periodE, periodN;
static int8_t    phaseA, phaseB, phaseC;
static int32_t   attackE, altE, holdE, haltE;
static int8_t    stepE, maxStepE;
static int16_t   outE, outN;
static uint32_t  lfsr;
static uint16_t  volA, volB, volC;
static uint16_t  *curVolA, *curVolB, *curVolC;

int8_t	  modeA, modeB, modeC;
int8_t	  toneA, toneB, toneC;
int8_t    noiseA, noiseB, noiseC;

char *ymSongTitle, *ymSongAuthor, *ymSongComment;


void ym_emu_init(const unsigned char *data) {
	uint32_t  i;
	uint32_t sampleBytes;

	ymData = data;

	numDigiDrums = ((uint16_t)ymData[0x14]) << 8;
	numDigiDrums |= ymData[0x15];

	digiDrumFormat = DIGI_DRUM_U8;
	if (ymData[0x13] & 0x04)
		digiDrumFormat = DIGI_DRUM_U4;
	else if (ymData[0x13] & 0x02)
		digiDrumFormat = DIGI_DRUM_S8;

	ymRegStream = (uint8_t*)ymData + 0x22;

	i = 0;
	while (numDigiDrums) {
		sampleBytes = ((uint32_t)ymRegStream[0]) << 24;
		sampleBytes |= ((uint32_t)ymRegStream[1]) << 16;
		sampleBytes |= ((uint32_t)ymRegStream[2]) << 8;
		sampleBytes |= ((uint32_t)ymRegStream[3]);
		ymRegStream += 4;
		if (i < 32) {
			digiDrumPtr[i] = ymRegStream;
			digiDrumLen[i] = sampleBytes;
			i++;
		}
		ymRegStream += sampleBytes;
		numDigiDrums--;
	}

	ymSongTitle = ymRegStream;
	while (*ymRegStream++);		// Skip song name
	ymSongAuthor = ymRegStream;
	while (*ymRegStream++);		// Skip author name
	ymSongComment = ymRegStream;
	while (*ymRegStream++);		// Skip song comment


	speedTN = (64000000 / SAMPLE_RATE) * 256;
	speedE = speedTN;
	maxStepE = 31;
	speedT1 = (2457600 * 512) / SAMPLE_RATE;
	ymEnvTable = YM2149_VOL_TB;

	if (ymData[0x17] == 0x1B) {
		// This is probably a ZX Spectrum tune with an AY clock of 1773400 Hz
		// (0x001B0F58 big-endian). Adjust the emulation speed accordingly.
		speedTN = (56748800 / SAMPLE_RATE) * 256;
		speedE = speedTN >> 1;
		maxStepE = 15;
		ymEnvTable = YM_VOL_TB;
	}

	frame = 0;
	lfsr = 0x10000;
	phaseA = phaseB = phaseC = 0;

	numFrames = (uint32_t)ymData[0x0E] << 8;
	numFrames += ymData[0x0F];

	loopOffs = (uint32_t)ymData[0x1C] << 8;
	loopOffs += ymData[0x1D];
	loopFrame = (uint32_t)ymData[0x1E] << 8;
	loopFrame += ymData[0x1F];

	effects[0].typ = SFX_NONE;
	effects[0].timerPeriod = 0;
	effects[0].channel = 0;

	effects[1].typ = SFX_NONE;
	effects[1].timerPeriod = 0;
	effects[1].channel = 0;

	digiDrumOverride[0] = 0;
	digiDrumOverride[1] = 0;
	digiDrumOverride[2] = 0;

	// Each register frame is preceeded by a 16-bit mask that
	// specifies which register values that have changed this
	// frame.
	newRegValues = ((uint16_t)ymRegStream[1]) << 8;
	newRegValues |= ymRegStream[0];
	dataOffs = 2;

	for (i = 0; i < 16; i++) {
		if (newRegValues & (1 << i)) {
			ymRegs[i] = ymRegStream[dataOffs++];
		}
	}

	holdE = altE = attackE = 0;
	toneA = toneB = toneC = 1;
	noiseA = noiseB = noiseC = 1;
	modeA = modeB = modeC = 0;
}


static void set_enve_shape(uint32_t enveShape)
{
	if (enveShape & 4)
		attackE = maxStepE;
	else
		attackE = 0;

	if (enveShape & 8) {
		holdE = (enveShape & 1) ? maxStepE : 0;
		altE  = (enveShape & 2) ? maxStepE : 0;
	} else {
		holdE = maxStepE;
		altE = attackE;
	}
	stepE = maxStepE;
	outE = ymEnvTable[attackE ^ stepE];
}


static void set_level(uint32_t chn, uint32_t level)
{
	if (0 == chn) {
		modeA = level & 0x10;
		volA = YM_VOL_TB[level & 0x0F];
		curVolA = modeA ? &outE : &volA;

	} else if (1 == chn) {
		modeB = level & 0x10;
		volB = YM_VOL_TB[level & 0x0F];
		curVolB = modeB ? &outE : &volB;

	} else if (2 == chn) {
		modeC = level & 0x10;
		volC = YM_VOL_TB[level & 0x0F];
		curVolC = modeC ? &outE : &volC;
	}

	haltE = (modeA | modeB | modeC) ^ 0x10;
}


static void calc_tone_noise_masks(ymEffect *effect)
{
	toneA   = (ymRegs[YM_MIXER] & 1);
	toneB   = (ymRegs[YM_MIXER] & 2) >> 1;
	toneC   = (ymRegs[YM_MIXER] & 4) >> 2;

	noiseA  = (ymRegs[YM_MIXER] & 8)  >> 3;
	noiseB  = (ymRegs[YM_MIXER] & 16) >> 4;
	noiseC  = (ymRegs[YM_MIXER] & 32) >> 5;

	if (SFX_DIGI_DRUM == effect->typ && effect->timerPeriod) {
		if (1 == effect->channel) {
			toneA = noiseA = 1;
		} else if (2 == effect->channel) {
			toneB = noiseB = 1;
		} else if (3 == effect->channel) {
			toneC = noiseC = 1;
		}
	}
}


static inline void step_effect(ymEffect *effect, uint32_t effectNum)
{
	effect->timerCount += speedT1;
	if (effect->timerCount >= effect->timerPeriod) {
		effect->timerCount -= effect->timerPeriod;
		if (SFX_SID_VOICE == effect->typ) {
			effect->currentValue ^= effect->amplitude;
			set_level(effect->channel - 1, effect->currentValue);

		} else if (SFX_DIGI_DRUM == effect->typ) {
			if (DIGI_DRUM_U4 == digiDrumFormat) {
				if (effect->samplePos < digiDrumLen[effect->sampleNum]) {
					set_level(effect->channel - 1, (digiDrumPtr[effect->sampleNum])[effect->samplePos] & 0x0F);
					effect->samplePos++;
				} else {
					// The end of this digidrum sample has been reached; stop the effect
					effect->channelSticky = 0;
					effect->channel = 0;
					effect->timerPeriod = 0;
					effect->typ = SFX_NONE;
					calc_tone_noise_masks(effect);
				}
			} else {
				if (effect->samplePos < digiDrumLen[effect->sampleNum]) {
					set_level(effect->channel - 1, YM_DIGIDRUM_8TO4_TB[((digiDrumPtr[effect->sampleNum])[effect->samplePos]) & 0xFF]);
					effect->samplePos++;
				} else {
					// The end of this digidrum sample has been reached; stop the effect
					effect->channelSticky = 0;
					effect->channel = 0;
					effect->timerPeriod = 0;
					effect->typ = SFX_NONE;
					calc_tone_noise_masks(effect);
				}
			}
		} else if (SFX_SYNC_BUZZ == effect->typ) {
			set_enve_shape(effect->amplitude);
		}
	}
}


static void start_effect(ymEffect *effect, uint32_t effectNum)
{
	if (effect->channel) {
		if (SFX_SID_VOICE == effect->typ || SFX_SYNC_BUZZ == effect->typ) {
			uint32_t amplitude = (1 == effect->channel) ? ymRegs[YM_LEVEL_A] : ((2 == effect->channel) ? ymRegs[YM_LEVEL_B] : ymRegs[YM_LEVEL_C]);
			amplitude &= 0x0F;
			if (amplitude != effect->amplitude) {
				effect->amplitude = amplitude;
				effect->currentValue = amplitude;
			}
		} else if (SFX_DIGI_DRUM == effect->typ) {
			effect->sampleNum = (1 == effect->channel) ? ymRegs[YM_LEVEL_A] : ((2 == effect->channel) ? ymRegs[YM_LEVEL_B] : ymRegs[YM_LEVEL_C]);
			effect->sampleNum &= 0x1F;
			effect->samplePos = 0;
		} else {

		}
		effect->timerPeriod = TIMER_PRESCALER_TB[ymRegs[6 + (effectNum << 1)] >> 5] * (uint32_t)ymRegs[14 + effectNum];
	} else {
		if (SFX_DIGI_DRUM == effect->typ) {
			// Don't stop digidrum effects until the end of the sample has been reached
			effect->channel = effect->channelSticky;
		} else {
			effect->channelSticky = 0;
			effect->timerPeriod = 0;
			effect->typ = SFX_NONE;
		}
	}
	calc_tone_noise_masks(effect);
}


void ym_emu_run(uint32_t numSamples, int16_t *buffer)
{
	int32_t i;
	int16_t  outA, outB, outC, out;

	if (ymRegs[YM_ENVE_SHAPE] != 0xFF) {
		set_enve_shape(ymRegs[YM_ENVE_SHAPE]);
	}

	modeA   = ymRegs[YM_LEVEL_A] & 0x10;
	modeB   = ymRegs[YM_LEVEL_B] & 0x10;
	modeC   = ymRegs[YM_LEVEL_C] & 0x10;

	toneA   = (ymRegs[YM_MIXER] & 1);
	toneB   = (ymRegs[YM_MIXER] & 2) >> 1;
	toneC   = (ymRegs[YM_MIXER] & 4) >> 2;

	noiseA  = (ymRegs[YM_MIXER] & 8)  >> 3;
	noiseB  = (ymRegs[YM_MIXER] & 16) >> 4;
	noiseC  = (ymRegs[YM_MIXER] & 32) >> 5;

	volA    = YM_VOL_TB[(ymRegs[YM_LEVEL_A] & 0x0F)];
	volB    = YM_VOL_TB[(ymRegs[YM_LEVEL_B] & 0x0F)];
	volC    = YM_VOL_TB[(ymRegs[YM_LEVEL_C] & 0x0F)];

	periodA = ymRegs[0];
	periodA |= (uint32_t)(ymRegs[1] & 0x0F) << 8;
	periodA <<= 16;

	periodB = ymRegs[2];
	periodB |= (uint32_t)(ymRegs[3] & 0x0F) << 8;
	periodB <<= 16;

	periodC = ymRegs[4];
	periodC |= (uint32_t)(ymRegs[5] & 0x0F) << 8;
	periodC <<= 16;

	periodN = (uint32_t)(ymRegs[6] & 0x1F) << 16;

	periodE = ymRegs[11];
	periodE |= (uint32_t)(ymRegs[12]) << 8;
	periodE <<= 16;

	curVolA = modeA ? &outE : &volA;
	curVolB = modeB ? &outE : &volB;
	curVolC = modeC ? &outE : &volC;

	haltE = (modeA | modeB | modeC) ^ 0x10;

    if ('6' == ymData[2]) {
		// YM6 supports 4 different effects; two of which can be active at the same time.
		// r1[7:6] selects the type of effect to use for effect 1, and r1[5:4] selects the channel to apply it on.
		// r3[7:6] selects the type of effect to use for effect 2, and r3[5:4] selects the channel to apply it on.
        effects[0].channel = (ymRegs[1] >> 4) & 3;
        if (effects[0].channel) {
        	effects[0].typ = (ymRegs[1] >> 6) & 3;
			effects[0].channelSticky = effects[0].channel;
		}

        effects[1].channel = (ymRegs[3] >> 4) & 3;
        if (effects[1].channel) {
        	effects[1].typ = (ymRegs[3] >> 6) & 3;
			effects[1].channelSticky = effects[1].channel;
		}
	} else {
        // YM5 and below only support SID-Voice and Digidrum.
        // r1,r6,r14 controls the SID-voice effect, and r3,r8,r15 controls the Digidrum effect
        effects[0].channel = (ymRegs[1] >> 4) & 3;
        if (effects[0].channel) {
        	effects[0].typ = SFX_SID_VOICE;
			effects[0].channelSticky = effects[0].channel;
		}

        effects[1].channel = (ymRegs[3] >> 4) & 3;
        if (effects[1].channel) {
			if (effects[1].channel != effects[1].channelSticky && effects[1].typ == SFX_DIGI_DRUM) {
				digiDrumOverride[effects[1].channelSticky - 1] = 0;
				calc_tone_noise_masks(&effects[1]);
			}
        	effects[1].typ = SFX_DIGI_DRUM;
			effects[1].channelSticky = effects[1].channel;
		}
	}

	start_effect(&effects[0], 0);
	start_effect(&effects[1], 1);

	for (i = 0; i < numSamples; i++) {
		posA += speedTN;
		posB += speedTN;
		posC += speedTN;

		if (posA >= periodA) {
			posA -= periodA;
			phaseA ^= 1;
		}

		if (posB >= periodB) {
			posB -= periodB;
			phaseB ^= 1;
		}

		if (posC >= periodC) {
			posC -= periodC;
			phaseC ^= 1;
		}

		if (!haltE) {
			posE += speedE;
			if (posE >= periodE) {
				posE -= periodE;
				stepE--;
				if (stepE < 0) {
					attackE ^= altE;
					stepE &= (holdE ^ maxStepE);
					haltE = holdE;
				}
				outE = ymEnvTable[stepE ^ attackE];
			}
		}

		posN += speedTN;
		if (posN >= periodN) {
			posN -= periodN;
			outN = lfsr & 1;
			lfsr = (lfsr >> 1) | ((uint32_t)(outN ^ ((lfsr >> 3) & 1)) << 16);
		}

		if (effects[0].timerPeriod)
			step_effect(&effects[0], 0);
		if (effects[1].timerPeriod)
			step_effect(&effects[1], 1);

		outA = -((phaseA | toneA) & (outN | noiseA));
		outB = -((phaseB | toneB) & (outN | noiseB));
		outC = -((phaseC | toneC) & (outN | noiseC));

		outA &= *curVolA;
		outB &= *curVolB;
		outC &= *curVolC;

		outA += outB + outC;
		*buffer++ = outA;
	}

	frame++;
	if (frame >= numFrames) {
		frame = loopFrame;
		dataOffs = loopOffs;
	}

	newRegValues = ((uint16_t)ymRegStream[dataOffs+1]) << 8;
	newRegValues |= ymRegStream[dataOffs];
	dataOffs += 2;

	for (i = 0; i < 16; i++) {
		if (newRegValues & (1 << i)) {
			ymRegs[i] = ymRegStream[dataOffs++];
		}
	}
}
